/* Object Rexx convenience class for making IBM's Bean Scripting Framework (BSF)
   available in an OO-style as well: Java classes and Java objects appear as
   Object Rexx classes and objects and one can send Object Rexx messages to them.

   Object Rexx programmers should appreciate this little class ...
   ;)

   date:
          2003-02-09, rgf, no dimensional limits on arrays anymore,
                           makearray(), supplier() now created at the Java side, merely re-using

          2003-01-29, Augsburg-->Vienna (train), adding direct JVM-loading support,
                      BSF_PROXY, .BSF.cls (for accessing pre-registered Java class objects)

          2003-01-28, Augsburg, renamed "BSF_ARRAY" to "BSF_ARRAY_PROXY" to denote its purpose
          2001-04-22, 2001-05-02 (on the airplane, returning home
                                  from the 12th International Rexx Symposium),
          2001-05-07 (fixed bug in bsf:init, if directly used to instantiate data)
          2001-05-26 (made method "bsf.exit" available at the instance level)
          2001-06-02 (BSF: made method "bsf.sleep" available at the instance level and
                      as "sleep" at the class level;
                      BSF_ARRAY_PROXY: MAKEARRAY and SUPPLIER now only return entries with
                                 values as does Object Rexx with native arrays)

   version:     1.1.0

   authors:     (c) 2001-2003 Rony G. Flatscher, University of Augsburg, University of Essen,
                    WU (pronounced: "vey-uh") University Vienna
                         Peter Kalender (October 2000 - February 2001), University of Essen

  *  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>

*/

      -- load the BSF4Rexx functions and start a JVM, if necessary
if rxFuncQuery("BSF") = 1 then   -- BSF() support not loaded yet ?
do
   call rxFuncAdd "BsfLoadFuncs", "BSF4Rexx", "BsfLoadFuncs"
   .error~say( "BSF.cls: registering the BSF4Rexx functions ...")
   call BsfLoadFuncs
   .error~say( "         loading the Java Virtual Machine (JVM) ...")
   call BsfLoadJava
end

   -- ---rgf, initialization, make the pre-registered class objects accessible via
   -- ".bsf.cls": beware using these classes without studying their definition and purpose !
   .local~bsf.cls=.directory~new
   .bsf.cls~Class.JC        = .bsf_reference~new( 'Class.class'        )   -- a Java Object class
   .bsf.cls~Object.JC       = .bsf_reference~new( 'Object.class'       )   -- a Java Object class
   .bsf.cls~Method.JC       = .bsf_reference~new( 'Method.class'       )   -- a Java Object class

   .bsf.cls~String.JC       = .bsf_reference~new( 'String.class'       )   -- a Java Object class
   .bsf.cls~Array.JC        = .bsf_reference~new( 'Array.class'        )   -- a Java Object class

   .bsf.cls~boolean         = .bsf_reference~new( 'boolean.class'      )   -- primitive
   .bsf.cls~boolean.JC      = .bsf_reference~new( 'Boolean.class'      )   -- a Java Object class
   .bsf.cls~byte            = .bsf_reference~new( 'byte.class'         )   -- primitive
   .bsf.cls~byte.JC         = .bsf_reference~new( 'Byte.class'         )   -- a Java Object class
   .bsf.cls~char            = .bsf_reference~new( 'char.class'         )   -- primitive
   .bsf.cls~Character.JC    = .bsf_reference~new( 'Character.class'    )   -- a Java Object class
   .bsf.cls~double          = .bsf_reference~new( 'double.class'       )   -- primitive
   .bsf.cls~Double.JC       = .bsf_reference~new( 'Double.class'       )   -- a Java Object class
   .bsf.cls~float           = .bsf_reference~new( 'float.class'        )   -- primitive
   .bsf.cls~Float.JC        = .bsf_reference~new( 'Float.class'        )   -- a Java Object class
   .bsf.cls~int             = .bsf_reference~new( 'int.class'          )   -- primitive
   .bsf.cls~Integer.JC      = .bsf_reference~new( 'Integer.class'      )   -- a Java Object class
   .bsf.cls~long            = .bsf_reference~new( 'long.class'         )   -- primitive
   .bsf.cls~Long.JC         = .bsf_reference~new( 'Long.class'         )   -- a Java Object class
   .bsf.cls~short           = .bsf_reference~new( 'short.class'        )   -- primitive
   .bsf.cls~Short.JC        = .bsf_reference~new( 'Short.class'        )   -- a Java Object class
   .bsf.cls~void            = .bsf_reference~new( 'void.class'         )   -- primitive
   .bsf.cls~Void.JC         = .bsf_reference~new( 'Void.class'         )   -- a Java Object class


/* -------------------------------------------------------------
   routine which turns an array into an argument string as expected
   by the Java side, starting with the indicated element
   ------------------------------------------------------------- */
:: ROUTINE makeArgString
   use arg args

   tmpString    = ""
   do i=1 to args~items
      tmpString = tmpString ', "' || args[i] || '"'
   END
   return tmpString


/* -------------------------------------------------------------
   routine which checks whether data returned from the Java side
   represents a Java object; if so, an Object Rexx proxy is created
   and returned, else the received string value
   ------------------------------------------------------------- */
:: ROUTINE bsf.checkResult public
   parse arg data

   if data=.bsf~null_string then return .nil

        /* if an object was returned, it has been registered on the Java side;
           therefore reuse, if possible or, create an Object Rexx proxy of it;
           first check whether the name was constructed on the Java side        */
  if pos("@", data)>0 then      -- is this a beanName, created by Java?
  do
     parse var data before '@' after
     if before <> "" then
     do
        if datatype(after, "X") then    -- o.k. second part a hexadecimal value?
        do
           if pos(" ", after)=0 then    -- o.k. no space there?
           do
              -- create an Object Rexx proxy object for the Java object

              dimension = verify(data, "[")-1  -- is it a Java array?
              -- if dimension>0 & dimension<6 then
              if dimension>0 then   -- rgf, 2003-02-09, no upper limit anymore
                 return .bsf_array_reference~new(data, dimension) -- return a BSF_ARRAY proxy

              return .bsf_reference~new(data)   -- return a BSF_REFERENCE prox
           end
        end
     end
  end

  return data           -- assume the string of a primitive type




/* *************************************************************
   Class:       IBM's "Bean Scripting Framework"
   Purpose:     the proxy super-class for the Java classes
   ************************************************************* */
::CLASS BSF PUBLIC

/* -------------------------------------------------------------
   class method allows for creating a proxy class for a Java class
   which is made available to this process by placing it into the
   .local environment
   ------------------------------------------------------------- */
::METHOD import class
  USE ARG classname, java_path

  new_class = .BSF~subclass(classname)  -- create a subclass of .BSFClass

                -- create code for the INIT-method of the new class
  method_txt = 'self~init:super("' || java_path || '", ARG(1,"A"))'
  new_class~define('init', method_txt)  -- define the init-code with the class

  .local~setentry(classname, new_class) -- put new class into the local environment


/* -------------------------------------------------------------
   class method which merely initializes the class attribute "null_string"
   to its default value
   ------------------------------------------------------------- */
::METHOD init           CLASS
  EXPOSE null_string
  null_string=".NIL"

/* -------------------------------------------------------------
   class attribute method stores and allows access to the string
   which is used to indicat "null" for the Java side; upon return
   of values from Java, a Java "null" will be represented via
   Object Rexx' .nil object
   ------------------------------------------------------------- */
::METHOD null_string    ATTRIBUTE CLASS


/* -------------------------------------------------------------
   class method shuts down the Java Virtual Machine (JVM), which started
   this instance of Rexx, this causes Rexx to be shut down as well; the
   Java side waits "waitTime" msecs before terminating the JVM
   ------------------------------------------------------------- */
::METHOD exit class
  use arg exitValue, waitTime

  if \datatype(exitValue, "Whole Number") then return BSF("exit")
  if \datatype(waitTime,  "W"           ) then return BSF("exit", exitValue)
  return bsf.checkResult(BSF("exit", exitValue, waitTime))

/* -------------------------------------------------------------
   class method to have Java sleep for us "sleepTime" seconds; resolution
   to a msec is calculated, although the Java side may deviate up to 15/100 secs
   (we are not talking realtime here ;)
   ------------------------------------------------------------- */
::METHOD sleep class
  use arg sleepTime

  return bsf.checkResult(BSF("sleep", sleepTime))


/* -------------------------------------------------------------
   method to construct an instance (a "bean"); arguments are
   passed to BSF which uses them to create the bean instance
   from the given Java class
   ------------------------------------------------------------- */
::METHOD init
  USE ARG javaClass, args

        -- create an instance of javaClass and register it with BSF,
        -- leave beanName empty, this way we get a unique name from Java
--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(ARG(2, "Array")) ")"

  if args~class = .array then tmpStr=makeArgString(args)
                         else tmpSTr=makeArgString(ARG(2, "Array"))

--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(args) ")"
  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' tmpStr ")"

  interpret code        /* execute the code, calling bsf() with 'registerBean'  */

        -- assign beanName to object, use the Java supplied name
  self~ObjectName=beanName

-- say "---> BSF.init  : bean=["||beanName||"]"  -- 2003-01-05, ---rgf
  return self


/* -------------------------------------------------------------
   destructor method (called by Object Rexx when object gets
   destroyed) to unregister the bean from the BSF registry; this
   then allows the JVM to garbage collect that Java object as well
   ------------------------------------------------------------- */
::METHOD uninit
-- say "---> BSF.uninit: bean=["||self~objectname||"]" -- 2003-01-05, ---rgf

  return BSF("unregisterBean", self~objectname)



/* -------------------------------------------------------------
   unknown method (gets called by Object Rexx whenever an unknown
   message was sent to an instance of this class); forwards the
   message to BSF including all of the received arguments
   ------------------------------------------------------------- */
::METHOD unknown
  USE ARG name, args

-- say "BSF.cls::unknown, name:" name "for" "'"self~objectname"'" "class:" self~class

  code = 'call BSF "invoke", "' || self~objectname || '", "' || name || '"' makeArgString(args)
-- say "unknwon, code:" code
  interpret code        -- execute this dynamically created Rexx string
  if var("RESULT") then return bsf.checkResult(result)      -- check whether primitive type or Java object
                   else return .nil                     -- nothing came back, i.e. Java returned a true "null"




/* -------------------------------------------------------------
   the following methods implement the individual BSF-calls as
   methods, allowing for invoking them in the OO style with the
   help of messages;

   "registerBean" is not made available explicitly, it is
   implicitly part of the creation of a new proxy object (see
   "INIT" method above)

   "unregisterBean" is not made available explicitly, it is
   implicitly part of the destruction of objects (see the
   "UNINIT" methods above)

   "getStaticValue" is not made available to instances as they
   are able to get (and even set) public (static) fields via
   "getFieldValue" resp. "setFieldValue"
   ------------------------------------------------------------- */
--   instance methods

:: METHOD bsf.addEventListener  /* add event listener to the Java object*/
   use arg eventSetName, filter, text
   return BSF( "addEventListener", self~objectname, eventSetName, filter, text )

:: METHOD bsf.exit              /* exit Rexx                    */
   FORWARD MESSAGE ("EXIT") TO (self~class)

:: METHOD bsf.invoke
   parse arg methodName
   -- create UNKNOWN format (first argument is unknown method name)
   return self~unknown(methodName, arg(2, 'A'))


:: METHOD bsf.getFieldValue
   parse arg fieldName
   return bsf.checkResult( bsf("getFieldValue", self~objectname, fieldName) )


:: METHOD bsf.setFieldValue
   parse arg fieldName
   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || fieldName || ,
                 "'" makeArgString(arg(2, 'A')) ")"
   interpret code
   return res


:: METHOD bsf.getPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   return bsf.checkResult( bsf("getPropertyValue", self~objectname, propertyName, index))


:: METHOD bsf.setPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'" makeArgString(arg(3, 'A')) ")"
   interpret code
   return res



--   class  methods
:: METHOD     lookupBean                class
   parse arg beanName
   return bsf.checkResult( bsf("lookupBean", beanName) )

:: METHOD     pollEventText             class
   parse arg timeOut
   if \datatype(timeOut, "Whole") then return bsf.checkResult( bsf("pollEventText"))
   return bsf.checkResult( bsf("pollEventText", timeout))    -- with timeout

:: METHOD     getStaticValue            class
   parse arg className, fieldName
   return bsf.checkResult( bsf("getStaticValue", className, fieldName))

:: METHOD     postEventText              class
   parse arg eventText
   return bsf("postEventText", eventText)

:: METHOD     wrapArray                  class
   parse arg beanName
   return bsf.checkResult( bsf("wrapArray", beanName) )

:: METHOD     wrapEnumeration            class
   parse arg beanName
   return bsf.checkResult( bsf("wrapEnumeration", beanName) )

   -- set the string which from now on represents the Java "null" on both sides,
   -- the Rexx and the Java side
:: METHOD     setRexxNullString         class
   parse arg newValue
   self~null_string = newValue                  -- set class attribute to the new value
   return bsf("setRexxNullString", newValue)    -- set the Java side to the new value



/* *************************************************************
   Class:       BSF_REFERENCE
   Purpose:     the proxy class for objects returned by the Java side;
                objects of this class, if destroyed, cause via its superclass
                to remove one from the appropriate instance counter of the
                BSF registry
   Restriction: Should only be used by classes/routines in this file

   ************************************************************* */
::CLASS BSF_REFERENCE subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn deregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   use arg beanName
   self~ObjectName=beanName



/* *************************************************************
   Class:       BSF_REFERENCE
   Purpose:     allows to wrap up Java objects as Object Rexx objects;
                if instances are destroyed, then no derigstration from
                the BSF registry takes place
   restriction: no check is done, whether the given beanName is really
                in the BSF-registry
   ************************************************************* */
::CLASS BSF_PROXY subclass BSF PUBLIC

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   use arg beanName
   self~ObjectName=beanName

-- say "---> BSF_PROXY.init  : bean=["||beanName||"]" -- 2003-01-29, ---rgf

   -- make sure that BSF's UNINIT method does not run (it'll deregister
   -- the bean from the BSF registry
::METHOD uninit
  NOP




/* *************************************************************
   Class:       BSF_ARRAY
   Purpose:     the proxy class for array objects returned by the Java side
   ************************************************************* */
::CLASS BSF_ARRAY_REFERENCE     subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   expose items dimension indexArray   -- set access to object variable
   use arg beanName, dimension  -- assign "dimension" directly to exposed object variable

   self~ObjectName=beanName
   items=-1                     -- set and indicate, that not calculated as of yet
      -- "indexArray" allows for indexing at's and put's with a Java int array
   indexArray=bsf("invoke", .bsf.cls~array.jc, "NewInstance", "o", .bsf.cls~int, "i", dimension)
   indexArray=.bsf_array_proxy~new(indexArray, 1, dimension)


::METHOD dimension
  expose dimension dimensions

  if arg(1, "Omitted") then return dimension    -- return number of dimensions

  if VAR( dimensions ) = .false then self~getData  -- no data available, get it
  return dimensions[arg(1)]                     -- return entries in this dimension



      -- get the needed information from ArrayWrapper, make it available with Object Rexx attributes
::METHOD getData private
  expose dimensions items

      -- get the array wrapper of this array
  aw=bsf('wrapArray', self~objectname)
  items=bsf('getFieldValue', aw, 'items')       -- get calculated 'items' value
  dimArr=bsf('getFieldValue', aw, 'sizeOfDimension') -- get the Java array containing the dimensions

  dimensions=.array~new                         -- create an Object Rexx array
  do i=1 to bsf('arrayLength', dimArr)          -- get length of 'dimensions' array
     dimensions[i]=bsf('arrayAt', dimArr, i-1)  -- use Java indexing (starting with '0')
  end


  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java side
  -- The indices can be given individually in comma-separated form or in form of a Rexx array
::METHOD at
   expose indexArray
   use arg arg1

      -- adjust Object Rexx 1-based indexing to Java 0-based indexing
   if arg1~class=.array then
   do i=1 to arg1~items
      call bsf "arrayPut", indexArray, "int", arg1[i]-1, i-1
   end
   else -- individual enter the indices into the index-array
   do i=1 to arg()
      call bsf "arrayPut", indexArray, "int", arg(i)-1, i-1
   end
   return bsf.checkResult( bsf('arrayAt', self~objectName, indexArray))


  -- allow for referring to array elements as if it was an Object Rexx array
::METHOD "[]"
  FORWARD MESSAGE ("AT")        -- synonym for "AT", use its implementation



  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java side
  -- The indices can be given individually in comma-separated form or in form of a Rexx array
::METHOD put
   expose indexArray
   use arg type, value, tmpArr

   if value=.nil then value=.bsf~null_string -- indicate to the Java side a "null"

   if tmpArr~class<>.array then  -- if third argument is not an array, create one
      tmpArr=arg(3, "Array")     -- get rest of arguments as an array

   do i=1 to tmpArr~items        -- adjust indices to Java-style (0-based indices)
      call bsf "arrayPut", indexArray, "int", tmpArr[i]-1, i-1
   end

   call bsf 'arrayPut', self~objectName, type, value, indexArray



  -- allow for referring to array elements as if it was an Object Rexx array;
  -- using this syntax variant has the following implication: the valueType and the
  -- value itself must be given as *one* string value; the first word is the valueType
  -- the rest the value!    e.g.: xyz[1, 2, 3] = "String Hi, there!"
::METHOD "[]="
  -- synonym for "PUT", use its implementation
  expose dimension
  parse arg type value, .  -- for this variant extract the type from the rest of the value

  tmpArr=arg(2, "Array")   -- get the rest of the arguments in form of an array
  tmpArrItems=tmpArr~items
  newArr=.array~new
  newArr[1]=type           -- save type information
  newArr[2]=value          -- save value

  do i=3 to 9999           -- assign the index-arguments
     j=i-2
     if j>tmpArrItems then leave
     newArr[i]=tmpArr[j]
  end
  FORWARD MESSAGE ("PUT") ARGUMENTS (newArr) -- let the PUT-method do the work



-- calculates (once) and returns number of individual elements in array
::METHOD items
  expose items  -- set access to object variable
  if items=-1 then self~getData     -- not yet set, get the data from the ArrayWrapper
  return items



-- this method creates a supplier method
::METHOD  SUPPLIER

  aw=bsf('wrapArray', self~objectname)
   -- wrap up the Supplier object, such that it will be kept in the BSF registry as long
   -- as the Object Rexx BSF proxy exists
  return bsf.checkResult( bsf('invoke', aw, 'supplier', "boolean", .true) )



-- this method is needed by "DO ... OVER"
::METHOD MAKEARRAY

  aw=bsf('wrapArray', self~objectname)    -- wrap up the array
  arr=bsf('invoke', aw, 'makearray', "boolean", .false, "boolean", .false) -- get
  arr=bsf('arrayAt', arr, 0)              -- get the appropriate Java array

  rexxArr=.array~new
  do i=1 to bsf('arrayLength', arr)          -- get length of 'dimensions' array
     rexxArr[i]=bsf('arrayAt', arr, i-1)  -- use Java indexing (starting with '0')
  end
  return rexxArr



   -- wraps the given Java object and allows for using the array methods on it;
   -- if the reference is not used anymore, it will *not* cause the Java object to be de-registered
   -- used in the initialisation of "BSF_ARRAY" to get primitive int array (for the dimensions) without initialising them as standalone BSF-arrays
   --
::CLASS BSF_ARRAY_PROXY SUBCLASS BSF_ARRAY_REFERENCE
::METHOD init
  use arg beanName, dimension, items         -- name of Java object as registered in the BSF registry

  self~objectName=beanName
  self~dimension =dimension
  self~items     =items

::method dimension attribute
::method items     attribute

::METHOD uninit            -- intercept destructor message, don't remove BSF registry entry
  NOP




